<?php

namespace ImportWP\Pro\Cli;

use ImportWP\Common\Importer\ImporterManager;
use ImportWP\Common\Util\Logger;
use ImportWP\Container;

class Command
{
    public static function register()
    {
        \WP_CLI::add_command('importwp', 'ImportWP\Pro\Cli\Command');
    }

    public function import($args, $assoc_args)
    {
        Logger::setRequestType('cli');
        $assoc_args['action']      = 'importwp';
        $importer_id = intval(trim($args[0]));

        if ($importer_id <= 0) {
            \WP_CLI::error("You must provide an importer id.");
        }

        $user = uniqid('iwp', true);

        /**
         * @var ImporterManager $importer_manager
         */
        $importer_manager = Container::getInstance()->get('importer_manager');

        $importer_model = $importer_manager->get_importer($importer_id);

        $session = null;

        if (isset($assoc_args['start'])) {

            // Clear existing import
            delete_site_option('iwp_importer_config_' . $importer_model->getId());
            delete_site_option('iwp_importer_state_' . $importer_model->getId());
            delete_site_option('iwp_importer_lock_' . $importer_model->getId());
            delete_site_option('iwp_importer_lock_timestamp_' . $importer_model->getId());

            /**
             * @var \WPDB $wpdb
             */
            global $wpdb;

            $wpdb->query("DELETE FROM {$wpdb->options} WHERE option_name LIKE 'iwp\_importer\_state\_" . $importer_model->getId() . "\_%'");

            \WP_CLI::log("Starting Import");

            $datasource = $importer_model->getDatasource();
            $attachment_id = false;

            switch ($datasource) {
                case 'remote':
                    $raw_source = $importer_model->getDatasourceSetting('remote_url');
                    $source = apply_filters('iwp/importer/datasource', $raw_source, $raw_source, $importer_model);
                    $source = apply_filters('iwp/importer/datasource/remote', $source, $raw_source, $importer_model);
                    $attachment_id =  $importer_manager->remote_file($importer_model, $source, $importer_model->getParser());
                    break;
                case 'local':
                    $raw_source = $importer_model->getDatasourceSetting('local_url');
                    $source = apply_filters('iwp/importer/datasource', $raw_source, $raw_source, $importer_model);
                    $source = apply_filters('iwp/importer/datasource/local', $source, $raw_source, $importer_model);
                    $attachment_id =  $importer_manager->local_file($importer_model, $source);
                    break;
            }

            $importer_model =  $importer_manager->get_importer($importer_id);

            if (is_wp_error($attachment_id)) {

                \WP_CLI::error($attachment_id->get_error_message());
                return;
            }
        } else {
            \WP_CLI::log("Resume Import");

            // Get latest session and resume it.
            $config = get_site_option('iwp_importer_config_' . $importer_id);
            $session = $config['id'];
        }

        $progress_bar = null; //\WP_CLI\Utils\make_progress_bar( 'Generating users', $count );
        $current_section = null;
        $progress_counter = 0;

        $update_timestamp = function ($status) use (&$progress_bar, &$current_section, &$progress_counter) {

            $section = $status['section'];

            if (isset($status['progress'][$section]) && $status['progress'][$section]['end'] > 0) {

                $progress = $status['progress'][$section];

                if ($section !== $current_section) {

                    if ($progress_bar) {
                        $progress_bar->finish();
                    }

                    $progress_counter = 0;
                    $progress_bar = \WP_CLI\Utils\make_progress_bar(ucfirst($section) . ' Records', $progress['end'] - $progress['start']);
                    $current_section = $section;
                }

                for ($progress_counter; $progress_counter < $progress['current_row']; $progress_counter++) {
                    $progress_bar->tick();
                }
            }
        };

        do {
            add_action('iwp/importer/status/save', $update_timestamp);
            $state = $importer_manager->import($importer_model, $user, $session);
            remove_action('iwp/importer/status/save', $update_timestamp);

            // On first run the session is null, if we run it again it will invalidate the session
            if (is_null($session)) {
                $config = get_site_option('iwp_importer_config_' . $importer_id);
                $session = $config['id'];
            }
        } while (in_array($state['status'], ['running', 'timeout']));

        if ($state['status'] === 'error') {
            \WP_CLI::error('Import error: ' . $state['message']);
        }

        Logger::clearRequestType();
        \WP_CLI::success(sprintf("Import Complete - Inserts: %d, Updates: %d, Deletes: %d, Skipped: %d, Errors: %d", $state['stats']['inserts'], $state['stats']['updates'], $state['stats']['deletes'], $state['stats']['skips'], $state['stats']['errors']));
    }
}
